@@ -7,8 +7,9 @@ from .connectors import FolderConnector, FileHandleConnector, UrlConnector, regi
     get_connector, add_connector, get_source
 from .extractors import *
 from .model import DocumentMetadata, ContentNode, ContentFeature, SourceMetadata, \
-    Document
+    Document, LocalModelStore, RemoteStore, RemoteModelStore, DocumentStore
 from .pipeline import Pipeline, PipelineStatistics, PipelineContext
 from .sinks import InMemoryDocumentSink, FolderSink
 from .steps import NodeTagger, TextParser, RollupTransformer, TagsToKeyValuePairExtractor
-from .stores import JsonDocumentStore, TableDataStore, DictDataStore, DataStoreHelper, LocalDocumentStore
+from .stores import JsonDocumentStore, TableDataStore, DictDataStore, DataStoreHelper, LocalDocumentStore, \
+    RemoteDocumentStore
