@@ -1,5 +1,6 @@
 from __future__ import annotations
 
+import errno
 import json
 import logging
 import os
@@ -15,12 +16,39 @@ from rich import print
 from kodexa.connectors import get_source
 from kodexa.connectors.connectors import get_caller_dir, FolderConnector
 from kodexa.model import Document
+from kodexa.model.model import RemoteStore
 from kodexa.pipeline import PipelineContext, Pipeline, PipelineStatistics
 from kodexa.stores import TableDataStore
+from kodexa.stores.stores import LocalDocumentStore
+from appdirs import AppDirs
 
 logger = logging.getLogger('kodexa.platform')
 
 
+dirs = AppDirs("Kodexa", "Kodexa")
+
+
+def get_config():
+    path = os.path.join(dirs.user_config_dir, '.kodexa.json')
+    if os.path.exists(path):
+        with open(path, 'r') as outfile:
+            return json.load(outfile)
+    else:
+        return {'url': None, 'access_token': None}
+
+
+def save_config(config_obj):
+    path = os.path.join(dirs.user_config_dir, '.kodexa.json')
+    if not os.path.exists(os.path.dirname(path)):
+        try:
+            os.makedirs(os.path.dirname(path))
+        except OSError as exc:  # Guard against race condition
+            if exc.errno != errno.EEXIST:
+                raise
+    with open(path, 'w') as outfile:
+        json.dump(config_obj, outfile)
+
+
 class PipelineMetadataBuilder:
     """
     Build a metadata representation of the pipeline for passing to an instance of the
@@ -32,6 +60,15 @@ class PipelineMetadataBuilder:
 
     def build_steps(self, pipeline_metadata: Dict):
         pipeline_metadata.metadata.steps = []
+        pipeline_metadata.metadata.stores = []
+
+        for pipeline_store in self.pipeline.stores:
+
+            if isinstance(pipeline_store.store, RemoteStore):
+                pipeline_metadata.metadata.stores.append(
+                    {"name": pipeline_store.name, "ref": pipeline_store.store.get_ref()})
+            else:
+                raise Exception("Pipeline refers to a non-remote store, deployment of local stores is not supported")
 
         for idx, step in enumerate(self.pipeline.steps):
 
@@ -131,11 +168,15 @@ class KodexaPlatform:
 
     @staticmethod
     def get_access_token():
-        return os.getenv('KODEXA_ACCESS_TOKEN')
+        kodexa_config = get_config()
+        access_token = os.getenv('KODEXA_ACCESS_TOKEN')
+        return access_token if access_token is not None else kodexa_config['access_token']
 
     @staticmethod
     def get_url():
-        return os.getenv('KODEXA_URL', "https://platform.kodexa.com")
+        kodexa_config = get_config()
+        env_url = os.getenv('KODEXA_URL', None)
+        return env_url if env_url is not None else kodexa_config['url']
 
     @staticmethod
     def set_access_token(access_token):
@@ -180,6 +221,7 @@ class KodexaPlatform:
     @staticmethod
     def deploy_extension_from_uri(path, organisation_slug):
         url = f"{KodexaPlatform.get_url()}/api/extensionPacks/{organisation_slug}?uri={path}"
+        logger.info(f"Publishing extension pack to organization {organisation_slug}")
         response = requests.post(url,
                                  headers={"x-access-token": KodexaPlatform.get_access_token(),
                                           "content-type": "application/json"})
@@ -190,79 +232,118 @@ class KodexaPlatform:
             raise Exception("Unable to deploy new extension")
 
     @staticmethod
-    def deploy(slug: str, pipeline: Pipeline, name: str = "A new pipeline", description: str = "A Kodexa Pipeline",
-               example_urls=None, more_info_url: str = None, force_replace=False,
-               public=False):
+    def resolve_ref(ref: str):
+        org_slug = ref.split('/')[0]
+        slug = ref.split('/')[1].split(":")[0]
 
-        if example_urls is None:
-            example_urls = []
+        version = None
 
-        builder = PipelineMetadataBuilder(pipeline)
+        if len(ref.split('/')[1].split(":")) == 2:
+            version = ref.split('/')[1].split(":")[1]
 
-        if '/' not in slug:
-            logger.error("A slug must be valid, i.e. org_slug/pipeline_slug")
+        return [org_slug, slug, version]
+
+    @staticmethod
+    def deploy(ref: str, kodexa_object, name: str = "A new object", description: str = "A new Kodexa object",
+               options=None, public=False, force_replace=False):
+
+        if '/' not in ref:
+            logger.error("A ref must be valid, i.e. org_slug/pipeline_slug:version")
             raise Exception("Invalid slug")
 
-        logger.info(f"Deploying pipeline {slug}")
+        [organization_slug, object_slug, object_version] = KodexaPlatform.resolve_ref(ref)
+
+        if options is None:
+            options = {}
+
+        metadata_object = Dict()
+        metadata_object.orgSlug = organization_slug
+        metadata_object.publicAccess = public
+        metadata_object.slug = object_slug
+        metadata_object.version = object_version
+        metadata_object.name = name
+        metadata_object.description = description
+
+        object_url = None
+
+        if isinstance(kodexa_object, Pipeline):
+
+            object_url = "pipelines"
+
+            if "example_urls" not in options:
+                example_urls = []
+            else:
+                example_urls = options["example_urls"]
+
+            if "more_info_url" not in options:
+                more_info_url = None
+            else:
+                more_info_url = options["more_info_url"]
+
+            builder = PipelineMetadataBuilder(kodexa_object)
+
+            metadata_object.type = 'pipeline'
+            metadata_object.metadata.parameters = []
+            metadata_object.metadata.steps = []
+            metadata_object.exampleUrls = example_urls
+            metadata_object.moreInfoUrl = more_info_url
+
+            builder.build_steps(metadata_object)
+
+        elif isinstance(kodexa_object, LocalDocumentStore):
+            object_url = 'stores'
+            metadata_object.type = 'store'
+            metadata_object.contentType = 'DOCUMENT'
+        else:
+            raise Exception("Unknown object type, unable to deploy")
+
+        logger.info(f"Deploying {metadata_object.type} {ref}")
+
         access_token = KodexaPlatform.get_access_token_details()
-        logger.info(f"Using organization {access_token}")
-
-        new_pipeline = Dict()
-
-        organization_slug = slug.split('/')[0]
-        pipeline_slug = slug.split('/')[1]
-
-        new_pipeline.slug = pipeline_slug
-        new_pipeline.name = name
-        new_pipeline.type = 'pipeline'
-        new_pipeline.description = description
-        new_pipeline.metadata.parameters = []
-        new_pipeline.metadata.steps = []
-        new_pipeline.orgSlug = organization_slug
-        new_pipeline.publicAccess = public
-        new_pipeline.exampleUrls = example_urls
-        new_pipeline.moreInfoUrl = more_info_url
-
-        builder.build_steps(new_pipeline)
 
         if organization_slug not in access_token.organizationSlugs:
             logger.error(f"You do not have access to the organization {organization_slug}")
             raise Exception("Unable to deploy, no access to organization")
 
-        response = requests.get(f"{KodexaPlatform.get_url()}/api/pipelines/{organization_slug}/{pipeline_slug}",
+        if object_version:
+            url = f"{KodexaPlatform.get_url()}/api/{object_url}/{organization_slug}/{object_slug}/{object_version}"
+        else:
+            url = f"{KodexaPlatform.get_url()}/api/{object_url}/{organization_slug}/{object_slug}"
+
+        response = requests.get(url,
                                 headers={"x-access-token": KodexaPlatform.get_access_token(),
                                          "content-type": "application/json"})
 
         if response.status_code == 401:
-            logger.error("You do not have the permissions to access this pipeline")
+            logger.error(f"You do not have the permissions to access this {metadata_object.type}")
             raise Exception("Not authorized on pipeline")
 
         if response.status_code == 404:
-            logger.info("Pipeline doesn't exist, will publish")
-            response = requests.post(f"{KodexaPlatform.get_url()}/api/pipelines/{organization_slug}",
-                                     json=new_pipeline.to_dict(),
+            logger.info("Object doesn't exist, will deploy")
+            response = requests.post(f"{KodexaPlatform.get_url()}/api/{object_url}/{organization_slug}",
+                                     json=metadata_object.to_dict(),
                                      headers={"x-access-token": KodexaPlatform.get_access_token(),
                                               "content-type": "application/json"})
             if response.status_code == 200:
-                logger.info("Pipeline deployed")
+                logger.info("Deployed")
             else:
                 logger.error(response.text)
-                raise Exception("Unable to deploy new pipeline")
+                raise Exception("Unable to deploy")
         else:
-            logger.info("Pipeline already exists")
+            logger.info(f"{ref} already exists")
             if force_replace:
-                logger.info("Replacing pipeline")
-                response = requests.put(f"{KodexaPlatform.get_url()}/api/pipelines/{organization_slug}/{pipeline_slug}",
-                                        json=new_pipeline.to_dict(),
+                logger.info(f"Replacing {ref}")
+                response = requests.put(url,
+                                        json=metadata_object.to_dict(),
                                         headers={"x-access-token": KodexaPlatform.get_access_token(),
                                                  "content-type": "application/json"})
                 if response.status_code == 200:
-                    logger.info("Pipeline deployed")
+                    logger.info("Deployed")
                 else:
                     logger.error(response.text)
-                    raise Exception("Unable to deploy and replace existing pipeline")
+                    raise Exception("Unable to deploy and replace")
             else:
-                logger.info("Not updating")
+                logger.warning("Not updating")
                 return
 
     @staticmethod
@@ -277,11 +358,9 @@ class KodexaPlatform:
             raise Exception("Unable to list objects")
 
     @staticmethod
-    def undeploy(slug: str):
-        organization_slug = slug.split('/')[0]
-        pipeline_slug = slug.split('/')[1]
-
-        response = requests.delete(f"{KodexaPlatform.get_url()}/api/pipelines/{organization_slug}/{pipeline_slug}",
+    def undeploy(ref: str):
+        url_ref = ref.replace(':', '/')
+        response = requests.delete(f"{KodexaPlatform.get_url()}/api/pipelines/{url_ref}",
                                    headers={"x-access-token": KodexaPlatform.get_access_token(),
                                             "content-type": "application/json"})
         if response.status_code == 200:
@@ -360,6 +439,27 @@ class KodexaPlatform:
         except:
             print(f"\n:exclamation: Failed to delete {object_type_metadata['name']} [{sys.exc_info()[0]}]")
 
+    @classmethod
+    def login(cls, kodexa_url, username, password):
+        from requests.auth import HTTPBasicAuth
+        obj_response = requests.get(f"{kodexa_url}/api/account/me/token",
+                                    auth=HTTPBasicAuth(username, password),
+                                    headers={"content-type": "application/json"})
+        if obj_response.status_code == 200:
+            kodexa_config = get_config()
+            kodexa_config['url'] = kodexa_url
+            kodexa_config['access_token'] = obj_response.text
+            save_config(kodexa_config)
+            print("Logged in")
+        else:
+            print(f"Check your URL and password [{obj_response.status_code}]")
+
+    @classmethod
+    def get_server_info(cls):
+        return requests.get(f"{KodexaPlatform.get_url()}/api",
+                            headers={"x-access-token": KodexaPlatform.get_access_token(),
+                                     "content-type": "application/json"}).json()
+
 
 class RemoteSession:
     """
@@ -371,6 +471,11 @@ class RemoteSession:
         self.slug = slug
         self.cloud_session = None
 
+    def get_action_metadata(self, ref):
+        r = requests.get(f"{KodexaPlatform.get_url()}/api/actions/{ref}",
+                         headers={"x-access-token": KodexaPlatform.get_access_token()})
+        return r.json()
+
     def start(self):
         logger.info(f"Creating session {self.slug} ({KodexaPlatform.get_url()})")
         r = requests.post(f"{KodexaPlatform.get_url()}/api/sessions", params={self.session_type: self.slug},
@@ -380,17 +485,21 @@ class RemoteSession:
             logger.error("Unable to create session")
             logger.error(r.text)
             raise Exception("Unable to create a session, check your URL and access token")
+
         self.cloud_session = Dict(json.loads(r.text))
 
     def execution_action(self, document, options, attach_source):
         files = {}
         if attach_source:
+            logger.info("Attaching source")
             files["file"] = get_source(document)
+            files["file_document"] = document.to_msgpack()
         else:
             files["document"] = document.to_msgpack()
 
         data = {"options": json.dumps(options), "document_metadata_json": json.dumps(document.metadata)}
 
+        logger.info(f"Executing session {self.cloud_session.id}")
         r = requests.post(f"{KodexaPlatform.get_url()}/api/sessions/{self.cloud_session.id}/execute",
                           params={self.session_type: self.slug},
                           data=data,
@@ -474,7 +583,7 @@ class RemotePipeline:
     """
 
     def __init__(self, slug, connector, version=None, attach_source=True, parameters=None, auth=None):
-        logging.info(f"Initializing a new pipeline {slug}")
+        logger.info(f"Initializing a new pipeline {slug}")
 
         if isinstance(connector, Document):
             self.connector = [connector]
@@ -503,7 +612,7 @@ class RemotePipeline:
 
         :param sink: the sink for the pipeline
         """
-        logging.info(f"Setting sink {sink.get_name()} on {self.slug}")
+        logger.info(f"Setting sink {sink.get_name()} on {self.slug}")
         self.sink = sink
 
         return self
@@ -511,12 +620,12 @@ class RemotePipeline:
     def run(self):
         self.context.statistics = PipelineStatistics()
 
-        logging.info(f"Starting remote pipeline {self.slug}")
+        logger.info(f"Starting remote pipeline {self.slug}")
         cloud_session = RemoteSession("pipeline", self.slug)
         cloud_session.start()
 
         for document in self.connector:
-            logging.info(f"Processing {document}")
+            logger.info(f"Processing {document}")
             execution = cloud_session.execution_action(document, self.parameters, self.attach_source)
             execution = cloud_session.wait_for_execution(execution)
 
@@ -527,7 +636,7 @@ class RemotePipeline:
             self.context.statistics.processed_document(document)
 
             if self.sink:
-                logging.info(f"Writing to sink {self.sink.get_name()}")
+                logger.info(f"Writing to sink {self.sink.get_name()}")
                 try:
                     self.sink.sink(document)
                 except:
@@ -539,7 +648,7 @@ class RemotePipeline:
                     if self.context.stop_on_exception:
                         raise
 
-        logging.info(f"Completed pipeline {self.slug}")
+        logger.info(f"Completed pipeline {self.slug}")
 
         return self.context
 
@@ -626,7 +735,16 @@ class RemoteAction:
     def process(self, document, context):
         cloud_session = RemoteSession("service", self.slug)
         cloud_session.start()
-        execution = cloud_session.execution_action(document, self.options, self.attach_source)
+
+        logger.info(f"Loading metadata for {self.slug}")
+        action_metadata = cloud_session.get_action_metadata(self.slug)
+
+        requires_source = False
+        if 'requiresSource' in action_metadata['metadata']:
+            requires_source = action_metadata['metadata']['requiresSource']
+
+        execution = cloud_session.execution_action(document, self.options,
+                                                   self.attach_source if self.attach_source else requires_source)
         execution = cloud_session.wait_for_execution(execution)
 
         result_document = cloud_session.get_output_document(execution)
@@ -634,6 +752,11 @@ class RemoteAction:
 
         return result_document if result_document else document
 
+    def end_processing(self):
+
+        # TODO not yet implemented for remote steps
+        pass
+
     def to_configuration(self):
         """
         Returns a dictionary representing the configuration information for the step
