@@ -11,7 +11,9 @@ from typing import Dict, Type
 
 import requests
 
-from kodexa.model import Document, DocumentMetadata
+from kodexa.model import Document, DocumentMetadata, DocumentStore
+
+logger = logging.getLogger('kodexa.connectors')
 
 
 def get_caller_dir():
@@ -68,7 +70,6 @@ class FolderConnector:
                 document.source.original_filename = os.path.basename(self.files[self.index - 1])
                 document.source.original_path = self.path
                 document.source.connector = self.get_name()
-                document.source.mime_type = mimetypes.guess_type(self.files[self.index - 1])
 
                 # TODO we need to get the checksum and last_updated and created times
                 return document
@@ -120,7 +121,6 @@ class FileHandleConnector:
             document.source.original_filename = self.file
             document.source.original_path = os.path.basename(self.file)
             document.source.connector = self.get_name()
-            document.source.mime_type = mimetypes.guess_type(self.file)
 
             # TODO we need to get the checksum and last_updated and created times
             return document
@@ -186,7 +186,7 @@ def get_connectors():
 
 def get_connector(connector, options):
     if connector in registered_connectors:
-        logging.info("Getting registered connector")
+        logger.info(f"Getting registered connector {connector}")
         return registered_connectors[connector]
     else:
         logging.error(f"Unable to find connector {connector}")
@@ -203,6 +203,25 @@ def get_source(document):
     return connector.get_source(document)
 
 
+class DocumentStoreConnector(object):
+
+    def __init__(self, store: DocumentStore, subscription: str):
+        self.store = store
+        self.subscription = subscription
+        self.index = 0
+
+    def __iter__(self):
+        return self
+
+    def __next__(self):
+        if self.index >= self.store.count():
+            raise StopIteration
+        else:
+            uuid_value = self.store.list_objects()[self.index]['uuid']
+            self.index += 1
+            return self.store.get_by_uuid(uuid_value)
+
+
 add_connector(FolderConnector)
 add_connector(FileHandleConnector)
 add_connector(UrlConnector)
