@@ -1,10 +1,13 @@
 import dataclasses
+import inspect
 import itertools
 import json
 import os
 import re
+import shutil
 import uuid
 from enum import Enum
+from pathlib import Path
 from typing import List, Optional, Any
 
 import msgpack
@@ -34,6 +37,20 @@ class Store:
         pass
 
 
+class RemoteStore:
+    """
+    A remote store is one that refers to a Kodexa platform  instance
+    """
+
+    def get_ref(self) -> str:
+        """
+        Get the reference to the store on the platform (i.e. kodexa/my-store:1.1.0)
+
+        :return: The reference
+        """
+        pass
+
+
 class DocumentMetadata(Dict):
     """
     A flexible dict based approach to capturing metadata for the document
@@ -149,7 +166,7 @@ class ContentNode(object):
         :rtype: ContentNode
         """
 
-        node_type = content_node_dict['type'] if document.version == Document.PREVOUS_VERSION else content_node_dict[
+        node_type = content_node_dict['type'] if document.version == Document.PREVIOUS_VERSION else content_node_dict[
             'node_type']
 
         new_content_node = document.create_node(node_type=node_type, content=content_node_dict[
@@ -1183,19 +1200,33 @@ class SourceMetadata:
     original_filename: Optional[str] = None
     original_path: Optional[str] = None
     checksum: Optional[str] = None
+
+    # The ID used for internal caching
+    cid: Optional[str] = None
     last_modified: Optional[str] = None
     created: Optional[str] = None
     connector: Optional[str] = None
     mime_type: Optional[str] = None
     headers: Optional[Dict] = None
 
+    # The UUID of the document that this document was derived from
+    # noting that multiple documents coming from an original source
+    lineage_document_uuid: Optional[str] = None
+
+    @classmethod
+    def from_dict(cls, env):
+        return cls(**{
+            k: v for k, v in env.items()
+            if k in inspect.signature(cls).parameters
+        })
+
 
 class Document(object):
     """
     A Document is a collection of metadata and a set of content nodes.
     """
 
-    PREVOUS_VERSION: str = "1.0.0"
+    PREVIOUS_VERSION: str = "1.0.0"
     CURRENT_VERSION: str = "2.0.0"
 
     def __str__(self):
@@ -1214,10 +1245,33 @@ class Document(object):
         self.version = Document.CURRENT_VERSION
         self.add_mixin('core')
         self.source: SourceMetadata = source
+        self.labels: List[str] = []
 
         # Make sure we apply all the mixins
         registry.apply_to_document(self)
 
+    def add_label(self, label: str):
+        """
+        Add a label to the document
+
+        :param label:str Label to add
+        :return: the document
+        """
+        if label not in self.labels:
+            self.labels.append(label)
+
+        return self
+
+    def remove_label(self, label: str):
+        """
+        Remove a label to the document
+
+        :param label:str Label to remove
+        :return: the document
+        """
+        self.labels.remove(label)
+        return self
+
     @classmethod
     def from_text(cls, text):
         new_document = Document()
@@ -1287,12 +1341,26 @@ class Document(object):
         :return: A dictionary representation of this Document.
         :rtype: dict
         """
+
+        # We don't want to store the none values
+        def clean_none_values(d):
+            clean = {}
+            for k, v in d.items():
+                if isinstance(v, dict):
+                    nested = clean_none_values(v)
+                    if len(nested.keys()) > 0:
+                        clean[k] = nested
+                elif v is not None:
+                    clean[k] = v
+            return clean
+
         return {'version': Document.CURRENT_VERSION, 'metadata': self.metadata,
                 'content_node': self.content_node.to_dict() if self.content_node else None,
-                'source': dataclasses.asdict(self.source),
+                'source': clean_none_values(dataclasses.asdict(self.source)),
                 'mixins': self._mixins,
                 'exceptions': self.exceptions,
                 'log': self.log,
+                'labels': self.labels,
                 'uuid': self.uuid}
 
     @staticmethod
@@ -1311,7 +1379,7 @@ class Document(object):
         for mixin in doc_dict['mixins']:
             registry.add_mixin_to_document(mixin, new_document)
         new_document.version = doc_dict['version'] if 'version' in doc_dict and doc_dict[
-            'version'] else Document.PREVOUS_VERSION  # some older docs don't have a version or it's None
+            'version'] else Document.PREVIOUS_VERSION  # some older docs don't have a version or it's None
         new_document.log = doc_dict['log'] if 'log' in doc_dict else []
         new_document.exceptions = doc_dict['exceptions'] if 'exceptions' in doc_dict else []
         new_document.uuid = doc_dict['uuid'] if 'uuid' in doc_dict else str(
@@ -1319,7 +1387,9 @@ class Document(object):
         if 'content_node' in doc_dict and doc_dict['content_node']:
             new_document.content_node = ContentNode.from_dict(new_document, doc_dict['content_node'])
         if 'source' in doc_dict and doc_dict['source']:
-            new_document.source = SourceMetadata(**doc_dict['source'])
+            new_document.source = SourceMetadata.from_dict(doc_dict['source'])
+        if 'labels' in doc_dict and doc_dict['labels']:
+            new_document.labels = doc_dict['labels']
         return new_document
 
     @staticmethod
@@ -1493,19 +1563,62 @@ class Document(object):
         else:
             return self.create_node(node_type='results')
 
+    def get_labels(self) -> List[str]:
+        """
+        Return the list of labels associated with this document
+
+        :return: list of associated labels
+        """
+        return self.labels
+
 
 class DocumentStore:
     """A document store supports storing, listing and retrieving Kodexa documents"""
 
-    def get(self, path: str) -> Document:
+    def get_by_uuid(self, uuid_value: str) -> Optional[Document]:
         pass
 
-    def list(self) -> List[Dict]:
+    def list_objects(self) -> List[Dict]:
+        pass
+
+    def list(self):
+        objects = self.list_objects()
+        self._draw_table(objects)
+
+    def query(self, query: str = "*"):
+        objects = self.query_objects(query)
+        self._draw_table(objects)
+
+    def _draw_table(self, objects):
+        from rich.table import Table
+        from rich import print
+
+        table = Table(title=f"Listing Objects")
+
+        cols = ['id', 'content_type', 'path']
+        for col in cols:
+            table.add_column(col)
+        for object_dict in objects:
+            row = []
+
+            for col in cols:
+                row.append(object_dict[col] if col in object_dict else '')
+            table.add_row(*row)
+
+        print(table)
+
+    def query_objects(self, query: str) -> List[Dict]:
         pass
 
     def put(self, path: str, document: Document):
         pass
 
+    def count(self) -> int:
+        return 0
+
+    def accept(self, document: Document):
+        return True
+
 
 class FileStore:
     """A file store supports storing, listing and retrieving native files"""
@@ -1523,8 +1636,62 @@ class FileStore:
 class ModelStore:
     """A model store supports storing and retrieving of a ML models"""
 
-    def get(self, path: str) -> Document:
+    def get(self, path: str):
         pass
 
-    def put(self, path: str, document: Document):
+    def put(self, path: str, content):
+        pass
+
+
+class LocalModelStore(ModelStore):
+
+    def __init__(self, store_path: str, force_initialize=False):
+        self.store_path = store_path
+        path = Path(store_path)
+
+        if force_initialize and path.exists():
+            shutil.rmtree(store_path)
+
+        if path.is_file():
+            raise Exception("Unable to load store, since it is pointing to a file?")
+        elif not path.exists():
+            path.mkdir(parents=True)
+
+    def to_dict(self):
+        return {
+            "type": "MODEL",
+            "data": {
+                "path": self.store_path
+            }
+        }
+
+    def get(self, object_path: str):
+        if Path(os.path.join(self.store_path, object_path)).is_file():
+            return open(os.path.join(self.store_path, object_path), 'rb')
+        else:
+            return None
+
+    def put(self, object_path: str, content):
+        path = Path(object_path)
+        with open(os.path.join(self.store_path, path), 'wb') as object_file:
+            object_file.write(content)
+
+
+class RemoteModelStore(ModelStore, RemoteStore):
+
+    def to_dict(self):
+        return {
+            "type": "MODEL",
+            "ref": self.ref
+        }
+
+    def __init__(self, ref: str):
+        self.ref = ref
+
+    def get(self, object_path: str):
+        # TODO implement
+        pass
+
+    def put(self, object_path: str, content):
+        # TODO implement
         pass
