@@ -4,12 +4,12 @@ import inspect
 import logging
 import os
 import sys
+import time
 import traceback
 import uuid
 from collections import KeysView
 from enum import Enum
 from inspect import signature
-from io import StringIO
 from textwrap import dedent
 from typing import List, Optional, Dict
 from uuid import uuid4
@@ -17,10 +17,12 @@ from uuid import uuid4
 import yaml
 
 from kodexa.connectors import FolderConnector
-from kodexa.connectors.connectors import get_caller_dir
+from kodexa.connectors.connectors import get_caller_dir, DocumentStoreConnector
 from kodexa.model import Document, Store
 from kodexa.stores.stores import DocumentStore
 
+logger = logging.getLogger('kodexa.pipeline')
+
 
 def new_id():
     return str(uuid.uuid4()).replace("-", "")
@@ -215,17 +217,17 @@ class PipelineStep:
         self.cache_path = cache_path
 
         if str(type(self.step)) == "<class 'type'>":
-            logging.info(f"Adding new step class {step.__name__} to pipeline {self.name}")
+            logger.info(f"Adding new step class {step.__name__} to pipeline")
             self.step = step
         elif callable(self.step):
-            logging.info(f"Adding new step function {step.__name__} to pipeline {self.name}")
+            logger.info(f"Adding new step function {step.__name__} to pipeline")
             self.name = step.__name__
         elif isinstance(self.step, str):
-            logging.info(f"Adding new remote step {step} to pipeline {self.name}")
+            logger.info(f"Adding new remote step {step} to pipeline")
             from kodexa import RemoteAction
             self.step = RemoteAction(step, options=options, attach_source=attach_source)
         else:
-            logging.info(f"Adding new step {step.get_name()} to pipeline {self.name}")
+            logger.info(f"Adding new step {step.get_name()} to pipeline")
 
     def to_dict(self):
         try:
@@ -255,6 +257,7 @@ class PipelineStep:
 
     def execute(self, context, document):
 
+        start = time.perf_counter()
         if self.cache_path:
             # Check to see if we have a cached file
             cache_name = self.get_cache_name(document)
@@ -267,9 +270,11 @@ class PipelineStep:
                 context.set_current_document(document)
 
                 result_document = None
-
+                logger.info(f"Starting step")
                 if str(type(self.step)) == "<class 'type'>":
 
+                    logger.info(f"Starting step based on class {self.step}")
+
                     # We need to handle the parameterization
                     import copy
 
@@ -305,14 +310,14 @@ class PipelineStep:
                         result_document = step_instance.process(document, context)
 
                 elif not callable(self.step):
-                    logging.info(f"Starting step {self.step.get_name()}")
+                    logger.info(f"Starting step {self.step.get_name()}")
 
                     if len(signature(self.step.process).parameters) == 1:
                         result_document = self.step.process(document)
                     else:
                         result_document = self.step.process(document, context)
                 else:
-                    logging.info(f"Starting step function {self.step.__name__}")
+                    logger.info(f"Starting step function {self.step.__name__}")
 
                     if len(signature(self.step).parameters) == 1:
                         result_document = self.step(document)
@@ -322,8 +327,12 @@ class PipelineStep:
                 if self.cache_path and result_document:
                     result_document.to_kdxa(self.get_cache_name(result_document))
 
+                end = time.perf_counter()
+                logger.info(f"Step completed (f{end - start:0.4f}s)")
+
                 return result_document
             except:
+                logger.warning("Step failed")
                 tt, value, tb = sys.exc_info()
                 document.exceptions.append({
                     "step": self.step.__name__ if callable(self.step) else self.step.get_name(),
@@ -354,6 +363,39 @@ class PipelineStep:
             if document.source.original_filename is not None else document.uuid
         return f"{self.cache_path}/{file_name}.kdxa"
 
+    def end_processing(self, context):
+        try:
+            self.step.end_processing(context)
+        except:
+            pass
+
+
+class LabelStep(object):
+    """
+    A simple step for handling the labelling for a document
+    """
+
+    def __init__(self, label: str, remove=False):
+        self.label = label
+        self.remove = remove
+
+    def get_name(self):
+        return f"Remove label {self.label}" if self.remove else f"Add label {self.label}"
+
+    def process(self, document: Document):
+        if self.remove:
+            document.remove_label(self.label)
+        else:
+            document.add_label(self.label)
+        return document
+
+
+class PipelineStore:
+
+    def __init__(self, name: str, store: Store):
+        self.name = name
+        self.store = store
+
 
 class Pipeline:
     """
@@ -370,8 +412,9 @@ class Pipeline:
     """
     context: PipelineContext
 
-    def __init__(self, connector, name="Default", stop_on_exception=True, logging_level=logging.INFO):
-        logging.info(f"Initializing a new pipeline {name}")
+    def __init__(self, connector, name: str = "Default", stop_on_exception: bool = True,
+                 logging_level=logger.info, apply_lineage: bool = True):
+        logger.info(f"Initializing a new pipeline {name}")
 
         if isinstance(connector, Document):
             self.connector = [connector]
@@ -379,13 +422,14 @@ class Pipeline:
             self.connector = connector
 
         self.steps: List[PipelineStep] = []
+        self.stores: List[PipelineStore] = []
         self.sink = None
         self.name = name
-        self.context: PipelineContext = PipelineContext()
-        self.context.stop_on_exception = stop_on_exception
+        self.stop_on_exception = stop_on_exception
         self.logging_level = logging_level
+        self.apply_lineage = apply_lineage
 
-    def add_store(self, name, store):
+    def add_store(self, name: str, store: Store):
         """
         Add the store to the pipeline so that it is available to the pipeline
 
@@ -395,8 +439,47 @@ class Pipeline:
         :param name: the name of the store (to refer to it)
         :param store: the store that should be added
         """
-        self.context.add_store(name, store)
+        self.stores.append(PipelineStore(name, store))
+        return self
 
+    def add_label(self, label: str, enabled=True, condition=None, options=None, attach_source=False,
+                  parameterized=False, cache_path=None):
+        """
+        Adds a label to the document
+
+        :param label: label to add
+        :param enabled: is the step enabled (default True)
+        :param condition: condition to evaluate before executing the step (default None)
+        :param options: options to be passed to the step if it is a simplified remote action
+        :param attach_source: if step is simplified remote action this determines if we need to add the source
+        :param parameterized: apply the pipeline's parameters to the options
+        :param cache_path: cache the document locally, note this is only for local pipelines
+        :return: the pipeline
+        """
+        self.steps.append(
+            PipelineStep(step=LabelStep(label), name=f"Add label {label}", enabled=enabled, condition=condition,
+                         options=options,
+                         attach_source=attach_source, parameterized=parameterized, cache_path=cache_path))
+        return self
+
+    def remove_label(self, label: str, enabled=True, condition=None, options=None, attach_source=False,
+                     parameterized=False, cache_path=None):
+        """
+        Adds a label to the document
+
+        :param label: label to remove
+        :param enabled: is the step enabled (default True)
+        :param condition: condition to evaluate before executing the step (default None)
+        :param options: options to be passed to the step if it is a simplified remote action
+        :param attach_source: if step is simplified remote action this determines if we need to add the source
+        :param parameterized: apply the pipeline's parameters to the options
+        :param cache_path: cache the document locally, note this is only for local pipelines
+        :return: the pipeline
+        """
+        self.steps.append(
+            PipelineStep(step=LabelStep(label, remove=True), name=f"Remove label {label}", enabled=enabled,
+                         condition=condition, options=options,
+                         attach_source=attach_source, parameterized=parameterized, cache_path=cache_path))
         return self
 
     def add_step(self, step, name=None, enabled=True, condition=None, options=None, attach_source=False,
@@ -446,18 +529,19 @@ class Pipeline:
 
         :param sink: the sink for the pipeline
         """
-        logging.info(f"Setting sink {sink.get_name()} on {self.name}")
+        logger.info(f"Setting sink {sink.get_name()} on {self.name}")
         self.sink = sink
 
         return self
 
-    def to_store(self, document_store: DocumentStore):
+    def to_store(self, document_store: DocumentStore, processing_mode: str = "update"):
         """
         Allows you to provide the sink store easily
 
         This will wrap the store in a document store sink
 
         :param document_store: document store to use
+        :param processing_mode: the processing mode (update or new)
         :return: the pipeline
         """
         from kodexa.sinks import DocumentStoreSink
@@ -497,21 +581,38 @@ class Pipeline:
         if parameters is None:
             parameters = {}
 
+        self.context = PipelineContext()
+        self.context.stop_on_exception = self.stop_on_exception
+        for pipeline_store in self.stores:
+            self.context.add_store(pipeline_store.name, pipeline_store.store)
+
         self.context.statistics = PipelineStatistics()
         self.context.parameters = parameters
 
-        logging.info(f"Starting pipeline {self.name}")
+        logger.info(f"Starting pipeline {self.name}")
         for document in self.connector:
 
-            log_stream = StringIO()
-            logging.basicConfig(stream=log_stream, level=self.logging_level)
-            logging.info(f"Processing {document}")
+            logger.info(f"Processing {document}")
+
+            initial_source_metadata = document.source
+            lineage_document_uuid = document.uuid
+
+            if self.sink:
+                if not self.sink.accept(document):
+                    logger.info("Skipping document, since sink won't accept")
+                    break
+
             for step in self.steps:
                 document = step.execute(self.context, document)
 
             if document:
+
+                if self.apply_lineage:
+                    document.source = initial_source_metadata
+                    document.source.lineage_document_uuid = lineage_document_uuid
+
                 if self.sink:
-                    logging.info(f"Writing to sink {self.sink.get_name()}")
+                    logger.info(f"Writing to sink {self.sink.get_name()}")
                     try:
                         self.sink.sink(document)
                     except:
@@ -523,17 +624,31 @@ class Pipeline:
                         if self.context.stop_on_exception:
                             raise
 
-                document.log = log_stream.getvalue()
-
                 self.context.statistics.processed_document(document)
                 self.context.output_document = document
             else:
-                logging.warning("A step did not return a document?")
+                logger.warning("A step did not return a document?")
 
-        logging.info(f"Completed pipeline {self.name}")
+        logger.info(f"Completing pipeline {self.name}")
+
+        for step in self.steps:
+            step.end_processing(self.context)
+
+        logger.info(f"Completed pipeline {self.name}")
 
         return self.context
 
+    @staticmethod
+    def from_store(store: DocumentStore, subscription=None):
+        """
+        Build a new pipeline with the input documents from a document store
+
+        :param store:DocumentStore The URL ie. https://www.google.com
+        :param subscription:str The subscription query to use
+        :return: A new instance of a pipeline
+        """
+        return Pipeline(DocumentStoreConnector(store, subscription))
+
     @staticmethod
     def from_url(url, headers=None):
         """
