@@ -10,8 +10,9 @@ from typing import List, Dict, Optional
 import requests
 
 from kodexa.model import Document, Store, DocumentStore
+from kodexa.model.model import RemoteStore, LocalModelStore, RemoteModelStore
 
-logger = logging.getLogger('kodexa-stores')
+logger = logging.getLogger('kodexa.stores')
 
 
 class JsonDocumentStore(Store):
@@ -35,14 +36,14 @@ class JsonDocumentStore(Store):
         if path.is_file():
             raise Exception("Unable to load store, since it is pointing to a file?")
         elif not path.exists():
-            logging.info(f"Creating new store in {store_path}")
+            logger.info(f"Creating new store in {store_path}")
             path.mkdir(parents=True)
 
             # Create an empty index file
             open(os.path.join(path, 'index.idx'), 'a').close()
         self.read_index()
 
-        logging.info(f"Found {len(self.document_ids)} documents in {store_path}")
+        logger.info(f"Found {len(self.document_ids)} documents in {store_path}")
 
     def get_name(self):
         return f"JSON Document Store [{self.store_path}]"
@@ -53,6 +54,9 @@ class JsonDocumentStore(Store):
     def __iter__(self):
         return self
 
+    def accept(self, document: Document):
+        return True
+
     def __next__(self):
         if self.index > len(self.document_ids) - 1:
             raise StopIteration
@@ -183,7 +187,7 @@ class TableDataStore(Store):
         """
 
         return {
-            "type": "table",
+            "type": "TABLE",
             "data": {
                 "columns": self.columns,
                 "rows": self.rows,
@@ -268,7 +272,7 @@ class DictDataStore:
         :rtype: dict
         """
         return {
-            "type": "dictionary",
+            "type": "DICTIONARY",
             "data": {
                 "dicts": self.dicts
             }
@@ -323,12 +327,22 @@ class DataStoreHelper:
         """
 
         if 'type' in dict:
-            if 'table' == dict['type']:
+            if 'TABLE' == dict['type']:
                 columns = dict['data']['columns'] if 'columns' in dict['data'] else None
                 rows = dict['data']['rows'] if 'rows' in dict['data'] else None
                 return TableDataStore(columns=columns, rows=rows)
-            elif 'dictionary' == dict['type']:
+            elif 'DICTIONARY' == dict['type']:
                 return DictDataStore(dict['data']['dicts'])
+            elif 'DOCUMENT' == dict['type']:
+                if 'ref' in dict:
+                    return RemoteDocumentStore(dict['ref'])
+                else:
+                    return LocalDocumentStore(dict['data']['path'])
+            elif 'MODEL' == dict['type']:
+                if 'ref' in dict:
+                    return RemoteModelStore(dict['ref'])
+                else:
+                    return LocalModelStore(dict['data']['path'])
             else:
                 return None
         else:
@@ -338,10 +352,17 @@ class DataStoreHelper:
 
 class LocalDocumentStore(DocumentStore):
 
-    def __init__(self, store_path: str, force_initialize: bool = False):
+    def __init__(self, store_path: str, force_initialize: bool = False, mode: str = 'ALL'):
+
+        modes = ['ALL', 'ONLY_NEW']
+
+        if mode not in modes:
+            raise Exception(f"LocalDocumentStore mode must be one of {','.join(modes)}")
+
         self.store_path = store_path
         self.index = 0
         self.metastore: List[Dict] = []
+        self.mode = mode
 
         path = Path(store_path)
 
@@ -351,7 +372,7 @@ class LocalDocumentStore(DocumentStore):
         if path.is_file():
             raise Exception("Unable to load store, since it is pointing to a file?")
         elif not path.exists():
-            logging.info(f"Creating new local document store in {store_path}")
+            logger.info(f"Creating new local document store in {store_path}")
             path.mkdir(parents=True)
 
             # Create an empty index file
@@ -360,13 +381,27 @@ class LocalDocumentStore(DocumentStore):
 
         self.read_metastore()
 
-        logging.info(f"Found {len(self.metastore)} documents in {store_path}")
+        logger.info(f"Found {len(self.metastore)} documents in {store_path}")
+
+    def to_dict(self):
+        return {
+            "type": "DOCUMENT",
+            "data": {
+                "path": self.store_path
+            }
+        }
+
+    def accept(self, document: Document):
+        if self.mode == 'ALL':
+            return True
+        if self.mode == 'ONLY_NEW':
+            return not self.exists(document)
 
     def read_metastore(self):
         """
-        Read the metadata stire
+        Read the metadata store
         """
-        self.metastore:List[Dict] = []
+        self.metastore: List[Dict] = []
         with open(os.path.join(self.store_path, 'metastore.json')) as f:
             self.metastore = json.load(f)
 
@@ -377,36 +412,115 @@ class LocalDocumentStore(DocumentStore):
         with open(os.path.join(self.store_path, 'metastore.json'), 'w') as f:
             json.dump(self.metastore, f)
 
-    def get(self, path: str) -> Document:
-        return Document.from_kdxa(os.path.join(self.store_path, path) + ".kdxa")
+    def get_by_uuid(self, uuid: str) -> Optional[Document]:
+        for metadata in self.metastore:
+            if metadata['uuid'] == uuid:
+                return Document.from_kdxa(os.path.join(self.store_path, metadata['uuid']) + ".kdxa")
+        return None
 
-    def list(self) -> List[Dict]:
+    def get_by_path(self, path: str) -> Optional[Document]:
+        for metadata in self.metastore:
+            if metadata['path'] == path:
+                return Document.from_kdxa(os.path.join(self.store_path, metadata['uuid']) + ".kdxa")
+        return None
+
+    def list_objects(self) -> List[Dict]:
         return self.metastore
 
+    def count(self) -> int:
+        return len(self.metastore)
+
+    def load_kdxa(self, path: str):
+        document = Document.from_kdxa(path)
+        self.put(document.uuid, document)
+
     def put(self, path: str, document: Document):
-        document.to_kdxa(os.path.join(self.store_path, path) + ".kdxa")
-        self.metastore.append({'metadata':document.metadata,'source': dataclasses.asdict(document.source),'uuid':document.uuid, 'path': path})
+        document.to_kdxa(os.path.join(self.store_path, document.uuid) + ".kdxa")
+        self.metastore.append(
+            {'metadata': document.metadata, 'source': dataclasses.asdict(document.source), 'uuid': document.uuid,
+             'id': document.uuid,
+             'content_type': 'Document',
+             'path': path, 'labels': document.get_labels()})
         self.write_metastore()
 
+    def exists(self, document):
+        """
+        Look to see if we have document with the same original path and original filename
 
-class RemoteDocumentStore(DocumentStore):
+        :param document: document to check
+        :return: True if it is already in the document store
+        """
+        for metadata in self.metastore:
+            if document.source.original_path == metadata['source']['original_path'] and \
+                    document.source.original_filename == metadata['source']['original_filename']:
+                return True
+        return False
 
-    def __init__(self, ref: str, query: str = "*"):
-        self.ref = ref
-        self.query = query
+
+class RemoteDocumentStore(DocumentStore, RemoteStore):
+
+    def __init__(self, ref: str):
+        self.ref: str = ref
         self.objects: List[Dict] = []
         self.page = 1
 
-    def put(self, path: str, document: Document):
+    def to_dict(self):
+        return {
+            "type": "DOCUMENT",
+            "ref": self.ref
+        }
 
+    def get(self, document_id: str) -> Optional[Document]:
         from kodexa import KodexaPlatform
+        doc = requests.get(
+            f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents/{document_id}",
+            headers={"x-access-token": KodexaPlatform.get_access_token()})
+        if doc.status_code == 200:
+            return Document.from_msgpack(doc.content)
+        elif doc.status_code == 404:
+            return None
+        else:
+            logger.error("Get document failed [" + doc.text + "], response " + str(doc.status_code))
+            raise Exception("Get document failed [" + doc.text + "], response " + str(doc.status_code))
 
-        files = {"file": document.to_msgpack(),"path": path}
-        content_object_response = requests.post(
-            f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents",
-            headers={"x-access-token": KodexaPlatform.get_access_token()}, files=files)
+    def delete(self, document_id: str):
+        from kodexa import KodexaPlatform
+        delete_document_response = requests.delete(
+            f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents/{document_id}",
+            headers={"x-access-token": KodexaPlatform.get_access_token()})
 
+        if delete_document_response.status_code == 200:
+            logger.info(f"Deleted document {document_id}")
+        else:
+            logger.error("Delete document failed [" + delete_document_response.text + "], response " + str(
+                delete_document_response.status_code))
+            raise Exception("Delete document failed [" + delete_document_response.text + "], response " + str(
+                delete_document_response.status_code))
+
+    def get_ref(self) -> str:
+        return self.ref
+
+    def get_by_path(self, path) -> Optional[Document]:
+        hits = self.query(f"path:{path}")
+        if len(hits) == 1:
+            return self.get(hits[0]['id'])
+        else:
+            return None
+
+    def put(self, path: str, document: Document):
+        from kodexa import KodexaPlatform
         try:
+            import io
+
+            logger.info(f"Putting document with path {path}")
+
+            files = {"file": document.to_msgpack()}
+            data = {"path": path}
+            content_object_response = requests.post(
+                f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents",
+                headers={"x-access-token": KodexaPlatform.get_access_token()},
+                files=files, data=data)
+
             if content_object_response.status_code == 200:
                 from addict import Dict
                 content_object = Dict(json.loads(content_object_response.text))
@@ -416,8 +530,8 @@ class RemoteDocumentStore(DocumentStore):
                 raise Exception("Execution creation failed [" + content_object_response.text + "], response " + str(
                     content_object_response.status_code))
         except JSONDecodeError:
-            logger.error("Unable to handle response [" + content_object_response.text + "], response " + str(
-                content_object_response.status_code))
+            logger.error(
+                f"Unable to handle response [{content_object_response.text}], response {content_object_response.status_code}")
             raise
 
     def get_next_objects(self):
@@ -433,6 +547,27 @@ class RemoteDocumentStore(DocumentStore):
         else:
             self.objects = content_objects_response.json()['content']
 
+    def query_objects(self, query: str) -> List[Dict]:
+        from kodexa import KodexaPlatform
+        list_content = requests.get(
+            f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents",
+            params={'query': query},
+            headers={"x-access-token": KodexaPlatform.get_access_token()})
+        if list_content.status_code != 200:
+            raise Exception(
+                f"Exception occurred while trying to fetch objects [{list_content.status_code}]")
+        return list_content.json()['content']
+
+    def list_objects(self) -> List[Dict]:
+        from kodexa import KodexaPlatform
+        list_content = requests.get(
+            f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents",
+            headers={"x-access-token": KodexaPlatform.get_access_token()})
+        if list_content.status_code != 200:
+            raise Exception(
+                f"Exception occurred while trying to fetch objects [{list_content.status_code}]")
+        return list_content.json()['content']
+
     def __iter__(self):
         return self
 
@@ -443,11 +578,7 @@ class RemoteDocumentStore(DocumentStore):
             content_object = self.objects.pop(0)
 
             if content_object['content_type'] == "Document":
-                from kodexa import KodexaPlatform
-                doc = requests.get(
-                    f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents/{content_object['id']}",
-                    headers={"x-access-token": KodexaPlatform.get_access_token()})
-                return Document.from_msgpack(doc.content)
+                return self.get(content_object['id'])
             else:
 
                 # TODO we need the connector?
